/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

#ifndef __MATRIX_SKELETONS_WITH_FUSION_TPP__
#define __MATRIX_SKELETONS_WITH_FUSION_TPP__

#include "matrix_skeletons_common.tpp"

#endif // __MATRIX_SKELETONS_WITH_FUSION_TPP__

namespace sketo {
  namespace _impl {

    //------------------------------------------------------------
    // Rules for Owerwriting
    
    template <typename F>
    template <typename T>
    void
    ms_genobj<F>::check_overwrite(_impl::dist_matrix_buffer<T>*,
                                  int *,
                                  bool *,
                                  bool *,
                                  bool *scan_shift) const
    {
      *scan_shift = false;
    }

    template <typename F, typename AS>
    template <typename T>
    void
    ms_mapobj<F, AS>::check_overwrite(_impl::dist_matrix_buffer<T>* target,
                                      int *refcount,
                                      bool *no_deps,
                                      bool *child_no_deps,
                                      bool *scan_shift) const
    {
      as.check_overwrite(target, refcount, no_deps, child_no_deps, scan_shift);
      *scan_shift = false;
    }

    template <typename F, typename AS, typename BS>
    template <typename T>
    void
    ms_zipwobj<F, AS, BS>::check_overwrite(_impl::dist_matrix_buffer<T>* target,
                                           int *refcount,
                                           bool *no_deps,
                                           bool *child_no_deps,
                                           bool *scan_shift) const
    {
      as.check_overwrite(target, refcount, no_deps, child_no_deps, scan_shift);
      bs.check_overwrite(target, refcount, no_deps, child_no_deps, scan_shift);
      *scan_shift = false;
    }
    
    
    template <typename OP, typename OT, typename AS, typename A>
    template <typename T>
    void ms_scanobj<OP, OT, AS, A>::check_overwrite(_impl::dist_matrix_buffer<T>* target,
                                                    int *refcount,
                                                    bool *no_deps,
                                                    bool *child_no_deps,
                                                    bool *scan_shift) const
    {
      int refcount2 = 0;
      bool no_deps2 = true;
      as.check_overwrite(target, &refcount2, &no_deps2, child_no_deps, scan_shift);
      *refcount += refcount2;
      *no_deps = *no_deps && no_deps2;
      *child_no_deps = no_deps2;
      *scan_shift = true;
    }

    template <typename OP, typename OT, typename AS, typename A>
    template <typename T>
    void ms_scanrobj<OP, OT, AS, A>::check_overwrite(_impl::dist_matrix_buffer<T>* target,
                                                     int *refcount,
                                                     bool *no_deps,
                                                     bool *child_no_deps,
                                                     bool *scan_shift) const
    {
      int refcount2 = 0;
      bool no_deps2 = true;
      as.check_overwrite(target, &refcount2, &no_deps2, child_no_deps, scan_shift);
      *refcount += refcount2;
      *no_deps = *no_deps && no_deps2;
      *child_no_deps = no_deps2;
      *scan_shift = true;
    }


    
    template <typename OP, typename OT, typename AS, typename A>
    dist_matrix<A>*
    ms_scanrobj<OP, OT, AS, A>::ow_impl(dist_matrix<A>* bs) const
    {
      return _impl::ms_adpt::scanrobj_ow_impl(bs, op, ot, as);
    }

    template <typename OP, typename OT, typename AS, typename A>
    dist_matrix<A>*
    ms_scanobj<OP, OT, AS, A>::ow_impl(dist_matrix<A>* bs) const
    {
      return _impl::ms_adpt::scanobj_ow_impl(bs, op, ot, as);
    }


    template <typename OP, typename OT, typename AS, typename A>
    void
    ms_adpt::scanobj_subeval(dist_matrix<A>* bs, const OP op, const OT ot, const AS as)
    {
      // common scan implementation
      _impl::ms_adpt::scan<OP, OT, A>(bs, op, ot, as);
    }

    template <typename OP, typename OT, typename AS, typename A>
    void
    ms_adpt::scanrobj_subeval(dist_matrix<A>* bs, const OP op, const OT ot, const AS as)
    {
      // common scanr implementation
      _impl::ms_adpt::scanr<OP, OT, A>(bs, op, ot, as);
    }

    template <typename OP, typename OT, typename AS, typename A>
    dist_matrix<A>*
    ms_adpt::scanobj_ow_impl(dist_matrix<A>* bs, const OP op, const OT ot, const AS as)
    {
      // scan ow implementation
      // this is the same as non-ow implementation for matrices.
      ms_adpt::scanobj_subeval(bs, op, ot, as);

      return bs;
    }

    template <typename OP, typename OT, typename AS, typename A>
    dist_matrix<A>*
    ms_adpt::scanrobj_ow_impl(dist_matrix<A>* bs, const OP op, const OT ot, const AS as)
    {
      // scanr ow implementation
      // this is the same as non-ow implementation for matrices.
      ms_adpt::scanrobj_subeval(bs, op, ot, as);

      return bs;
    }

    template <int di, int dj, typename F, typename AS, typename A>
    template <typename T>
    void
    ms_shiftobj<di, dj, F, AS, A>::check_overwrite(_impl::dist_matrix_buffer<T>* target,
					 int *refcount,
					 bool *no_deps,
					 bool *child_no_deps,
					 bool *scan_shift) const
    {
      int refcount2 = 0;
      bool no_deps2 = true;
      as.check_overwrite(target, &refcount2, &no_deps2, child_no_deps, scan_shift);
      *refcount += refcount2;
      *no_deps = *no_deps && no_deps2;
      *child_no_deps = no_deps2;
      *scan_shift = true;
    }

    template <int di, int dj, typename F, typename AS, typename A>
    dist_matrix<A>*
    ms_shiftobj<di, dj, F, AS, A>::ow_impl(dist_matrix<A>* bs) const
    {
      _impl::ms_adpt::shift_common<di, dj, F, AS, A>(bs, *si, f, as);
      return bs;
    }

  }
}
