#include <iostream>
#include <cmath>
#include <utility>

#include "testcase.h"

void make_vector(double *ap, double *ae, double *aw, double *an, double *as, double *ab, double *at,
		 double *b, double *f)
{
  for (int m = 0; m < nmax_b; m++) { b[m] = 0.0; f[m] = 0.0; }

  for (int k = 1; k <= kmax; k++) {
    for (int j = 1; j <= jmax; j++) {
      for (int i = 1; i <= imax; i++) {
	int m = i + j*(imax+2) + k*(imax+2)*(jmax+2);
	f[m] = 1.0;
	b[m] = 2.0;
	if ((i > 1) && (i < imax)) b[m] += 2.0; else b[m] += 1.0;
	if ((j > 1) && (j < jmax)) b[m] += 2.0; else b[m] += 1.0;
	if ((k > 1) && (k < kmax)) b[m] += 2.0; else b[m] += 1.0;
      }
    }
  }
  *ap = 2.0;
  *ae = *aw = *an = *as = *at = *ab = 1.0;
}

void probmv(double ap, double ae, double aw, double an, double as, double at, double ab,
	    double *b, double *c, double *f)
{
  const int dx = 1;
  const int dy = imax + 2;
  const int dz = (imax + 2) * (jmax + 2);
#pragma omp parallel for
  for (int m = (imax + 2) * (jmax + 2); m < nmax_b - (imax + 2) * (jmax + 2); m++) {
    c[m] = (ap * b[m] + aw * b[m-dx] + ae * b[m+dx]
	    + as * b[m-dy] + an * b[m+dy]
	    + ab * b[m-dz] + at * b[m+dz]) * f[m];
  }

}

std::pair<int, double>
bicgstab_band(double ap, double ae, double aw ,double an, double as, double at, double ab,
	      double *x, double *b, double *f)
{
//   double er0 = EPS;

  double *c = new double[nmax_b];
  probmv(ap, ae, aw, an, as, at, ab, x, c, f);

  double *r = new double[nmax_b];
  double c1 = 0;
  for (int m = 0; m < nmax_b; m++) { r[m] = b[m] - c[m]; c1 += r[m] * r[m]; }
//   if (c1 < er0) {
//     delete[] r;
//     delete[] c;
//     return std::pair<int, double>(0, c1);
//   }

  double *p  = new double[nmax_b]; for (int m = 0; m < nmax_b; m++) { p[m] = r[m]; }
  double *r0 = new double[nmax_b]; for (int m = 0; m < nmax_b; m++) { r0[m] = r[m]; }

  double *e  = new double[nmax_b];
  double *y  = new double[nmax_b];
  double *v  = new double[nmax_b];
  std::pair<int, double> ret;
  for (int it = 0; it < ITRMAX; it++) {
    probmv(ap, ae, aw, an, as, at, ab, p, y, f);
    
    double c2 = 0;
    for (int m = 0; m < nmax_b; m++) { c2 += r0[m] * y[m]; }
    double alp = c1 / c2;
    
    for (int m = 0; m < nmax_b; m++) { e[m] = r[m] - alp*y[m]; }

    probmv(ap, ae, aw, an, as, at, ab, e, v, f);
    
    double ev = 0;
    for (int m = 0; m < nmax_b; m++) { ev += e[m] * v[m]; }
    double vv = 0;
    for (int m = 0; m < nmax_b; m++) { vv += v[m] * v[m]; }
    
    double c3 = ev / vv;
    for (int m = 0; m < nmax_b; m++) { x[m] = x[m] + alp*p[m] + c3*e[m]; }
    for (int m = 0; m < nmax_b; m++) { r[m] = e[m] - c3*v[m]; }

    double rr = 0;
    for (int m = 0; m < nmax_b; m++) { rr += r[m] * r[m]; }
//     if (std::fabs(rr) < er0) { ret = std::pair<int, double>(it + 1, rr); break; }

    c1 = 0;
    for (int m = 0; m < nmax_b; m++) { c1 += r0[m] * r[m]; }
    double bet = c1/(c2 * c3);
    for (int m = 0; m < nmax_b; m++) { p[m] = r[m] + bet*(p[m]-c3*y[m]); }

    ret = std::pair<int, double>(it + 1, rr);
  }
  delete[] v;
  delete[] y;
  delete[] e;
  delete[] r0;
  delete[] p;
  delete[] r;
  delete[] c;
  return ret;
}
  
int main(int, char**)
{
  double ap, ae, aw, an, as, at, ab;
  double *b = new double[nmax_b];
  double *f = new double[nmax_b];
  make_vector(&ap, &ae, &aw, &an, &as, &at, &ab, b, f);

  double *x = new double[nmax_b];
  for (int m = 0; m < nmax_b; ++m) { x[m] = 0.0; }
  for (int k = 1; k <= kmax; k++) {
    for (int j = 1; j <= jmax; j++) {
      for (int i = 1; i <= imax; i++) {
	int m = i + j*(imax+2) + k*(imax+2)*(jmax+2);
	x[m] = 1.0 + 0.01;
      }
    }
  }

  clock_t st, et;
  st = clock();
#ifndef PROBMV_ONLY
  std::pair<int, double> result = bicgstab_band(ap, ae, aw, an, as, at, ab,
						x, b, f);
//   std::cout << "errer = " << result.second << std::endl;
#else
  ap = -4.5;
  double *y = new double[nmax_b];
  // avoid too-much optimization
  double tmpval = 0.0;
  for (int i = 0; i < ITRMAX * 2; i++) {
    probmv(ap, ae, aw, an, as, at, ab, x, y, f);
    tmpval += y[0];
  }
  std::cout << tmpval << std::endl;
#endif
  et = clock();
  std::cout << "time = " << (double)(et - st)/CLOCKS_PER_SEC << std::endl;
  
}
