/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/***
 * @file fnorm.cpp
 * @brief The Frobenius Norm.  
 * @author Kento EMOTO
 */
#include <cstdlib>
#include <iostream>
#include <cmath>
#include "sketo/matrix_skeletons.h"

class to_double_t : public sketo::functions::base <double(sketo::mindex)> {
public: double operator()(sketo::mindex mi) const { return mi.i / (mi.j + 1.0); }
} to_double;

class sqr_t : public sketo::functions::base <double(double)> {
public: double operator()(double x) const { return x * x; }
} sqr;

int sketo::main(int argc, char *argv[])
{
	int n = 4000;
	if(argc > 1){
		n = atoi(argv[1]);
	}
	int t = 0;
	if(argc > 2) t = atoi(argv[2]);

  double norm2 = sketo::matrix_skeletons::reduce(
    std::plus<double>(),
    std::plus<double>(),
    sketo::matrix_skeletons::map(
      sqr,
      sketo::matrix_skeletons::generate(msize(n, n), to_double)));
  
  double norm = sqrt(norm2);
    
	sketo::cout << "#procs = "<< sketo::procs << std::endl;
	sketo::cout << "matrix size = " << n << "x" << n << std::endl;
	sketo::cout << "The Frobenius Norm = " << norm << std::endl;
	return 0;
}
